/*
    uart programmer for Cypress PSoC.
    CYACD file loader.
    Copyright (C) 2015  Michael Stickel (michael@cubic.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "cypress_cyacd_file.h"
#include <ctype.h>

#include "tools.h"

#include <stdio.h>
#include <stdlib.h>


/*
 * counts the number of valid hex digits.
 * stops at the first non hex digit or at the end
 *
 */
static int countDigits(const char *_str)
{
    int count = 0;
    while (_str[count] && isxdigit(_str[count]))
    {
	++count;
    }
    return count;
}

/*
 * convert single char hex digits to an integer.
 * If an unexpected char is found, -1 is returned.
 */
static int hexdigittoint ( char _c)
{
    if ((_c >= '0') && (_c <= '9'))
    {
	return _c - '0';
    }
    else if ((_c >= 'a') && (_c <= 'f'))
    {
	return _c - 'a' + 10;
    }
    else if ((_c >= 'A') && (_c <= 'F'))
    {
	return _c - 'A' + 10;
    }
    return -1;
}

/*
 * return the integer representation of a hexadecimal string. Limit it to the specified size in characters.
 */
static unsigned long hextoul(const char * _str, const int _maxlen)
{
    unsigned long v = 0;
    int i;
    for (i = 0; _str[i] && (i < _maxlen); ++i)
    {
	const int r = hexdigittoint(_str[i]);
	if (r < 0)
	{
	    return 0;
	}
	v = v*16 + r;
    }
    return v;
}

/*
 * Decode one row of an cyacd file and fill the structure _row with the data decoded.
 */
static int PsocFileDecodeRow(FlashRow & _row, const char * _rowString)
{
    /*
     * The flash lines have the following format:
     * [1 byte array ID] [2 bytes row number] [2 bytes data length] [N bytes of data] [1 byte checksum]
     */

    if (_rowString[0] != ':')
    {
	return 1;
    }

    const int digitCount = countDigits(&_rowString[1]);
    if ((digitCount % 2) == 0)
    {
	_row.flashArrayId = hextoul(&_rowString[1], 2);
	_row.flashRowNumber = hextoul(&_rowString[1+2], 4);
	_row.byteCount = hextoul(&_rowString[1+2+4], 4);
	if (_row.byteCount > 0 && _row.byteCount <= 256)
	{
	    _row.checksum  = hextoul(&_rowString[1+2+4+4+2*_row.byteCount], 2);
	    for (int i=0; i < _row.byteCount; ++i)
	    {
		_row.data[i] = hextoul(&_rowString[1+2+4+4+2*i], 2);
	    }
	    return 0;
	}
    }
    else
    {
	fprintf (stderr, "digit count not even: %d %d\n", digitCount, digitCount % 2);
    }
    return 1;
}

/*
 * Load the data of a Cypress cyacd file and fill it into <_data>.
 */
int PsocFileLoadCyacd(const char * _filename, CyacdFileContent & _data)
{
    //@TODO : add more error checking
    FILE * f = fopen(_filename, "r");
    if (f)
    {
	/*
	 * The header of this file has the following format:
	 *
	 * [4 bytes Silicon ID] [1 byte Silicon rev] [1 byte checksum type]
	 *
	 * The checksum type in the header indicates the type of checksum used in the packets sent between
	 * the bootloader and the bootloader host during the bootloading operation. If this byte is 0,
	 * the checksum is a basic summation. If it is 1, the checksum is CRC-16.
	 */
	char header[20];
	if (fgets(header, sizeof(header), f))
	{
	    _data.siliconId = hextoul(header, 8);
	    _data.siliconRev = hextoul(&header[8], 2);
	    _data.checksumType = strtol(&header[10], 0, 16);
	}

	char lineBuffer[1+8+2*256+2 + 10];
	while (fgets(lineBuffer, sizeof(lineBuffer), f))
	{
	    FlashRow  row;
	    if (PsocFileDecodeRow(row, lineBuffer) == 0)
	    {
		_data.rows.push_back(row);
	    }
	}

	_data.firstRowToProgram = 0x10000;
	_data.lastRowToProgram = -1;
	for (std::vector<FlashRow>::const_iterator it = _data.rows.begin(); it != _data.rows.end(); it++)
	{
	    if (_data.firstRowToProgram > it->flashRowNumber)
	    {
		_data.firstRowToProgram = it->flashRowNumber;
	    }
	    if (_data.lastRowToProgram < it->flashRowNumber)
	    {
		_data.lastRowToProgram = it->flashRowNumber;
	    }
	}
	fclose(f);
	return 0;
    }
    return 1;
}

/*
 * Dump the data from the cyacd file content on standard output.
 */
void PsocFileDumpContents(const CyacdFileContent & _data, const int _dumlevel)
{
    if (_dumlevel > 0)
    {
	printf ("siliconId: %08lX\n", _data.siliconId);
	printf ("siliconRev: %d\n", _data.siliconRev);
	printf ("checksumType: %d - %s\n",
		_data.checksumType,
		(_data.checksumType==0) ? "basic sum" : (_data.checksumType==1) ? "CRC-16" : "???"
	    );
	printf ("rowCount: %zd\n", _data.rows.size());
    }

    if (_dumlevel > 1)
    {
	for (std::vector<FlashRow>::const_iterator it = _data.rows.begin(); it != _data.rows.end(); it++)
	{
	    printf ("%d.%d (size:%d) csum:%02X\n", 
		    it->flashArrayId,
		    it->flashRowNumber,
		    it->byteCount,
		    it->checksum);
	    if (_dumlevel > 2)
	    {
		
		hexdump(it->data, it->byteCount);
	    }
	}
    }
}
