/*
    uart programmer for Cypress PSoC.
    Copyright (C) 2015  Michael Stickel (michael@cubic.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include <stdio.h>

#include "psoc_uart_programmer_commands.h"
#include "cypress_cyacd_file.h"

const char * PsocProgrammerVersion = "0.1";


int PsocProgramFileData(CyacdFileContent & _data)
{
    for (std::vector<FlashRow>::const_iterator it = _data.rows.begin(); it != _data.rows.end(); it++)
    {
	printf ("programming flash-array: %d row:%d\n",
		it->flashArrayId,
		it->flashRowNumber);

	//-- use 16byte chunks
	//-- write all but the last up to 16 bytes
	int retries = 0;
	int remain = it->byteCount;
	const unsigned char * pdata = it->data;
	while (remain > 16)
	{
	    //@TODO : more precize error checking
	    if (PsocSendData(pdata, 16) == 0)
	    {
		retries = 0;
		remain -= 16;
		pdata += 16;
	    }
	    else
	    {
		++retries;
		if (retries > 10)
		{
		    fprintf (stderr, "timeout in transfer during programming\n");
		    return 1;
		}
	    }
	}

	//-- write the last up to 16 bytes together with the flash command.
	//@TODO : more precize error checking
	if (PsocProgramRow(it->flashArrayId, it->flashRowNumber, pdata, remain) != 0)
	{
	    fprintf (stderr, "error during programming of flash-array:%d row:%d\n",
		     it->flashArrayId,
		     it->flashRowNumber);
	}

	//-- now verify
	unsigned int programmedChecksum = 0;
	if (PsocVerifyRow(it->flashArrayId, it->flashRowNumber, &programmedChecksum) == 0)
	{
	    if (programmedChecksum != it->checksum)
	    {
		fprintf (stderr, "checksumm error for flash-array:%d row:%d -- checksum chip:%02X file:%02X\n",
			 it->flashArrayId,
			 it->flashRowNumber,
			 programmedChecksum,
			 it->checksum);
	    }
	}
    }

    if (PsocVerifyChecksum())
    {
	return 1;
    }

    return 0;
}

static void usage(const char * _program)
{
    printf ("\npsoc_uart_programmer version %s Copyright (C) 2015  Michael Stickel\n"
	    "This program comes with ABSOLUTELY NO WARRANTY; See gpl-3.0.txt for details.\n"
	    "This is free software, and you are welcome to redistribute it\n"
	    "under the conditions of the GPL.\n", PsocProgrammerVersion);
    printf ("\n%s <serial-port> [<cyacd-file>]\n\n", _program);
    printf ("Before you call this program, the baudrate needs to be set correctly. You can use stty to do so:\n");
    printf ("   stty -F <serial-port> <baudrate>\n\n");
    printf ("If called with only one parameter, it outputs the device-id, silicon-rev, bootloader version and the available min and max row-numbers\n\n");
}

int main (int argc, char **argv)
{
    if (argc <= 1)
    {
	usage(argv[0]);
	return 1;
    }

    if (PsocOpenPort(argv[1]) < 0)
    {
	fprintf (stderr, "failed to open serial port \"%s\"\n", argv[1]);
	return 1;
    }

    //---
    //--- connect to device
    //---

    unsigned long deviceId = 0;
    unsigned char siliconRev = 0;
    int bootloaderVersion = 0;
    if (PsocEnterBootloader (&deviceId, &siliconRev, &bootloaderVersion) == 0)
    {
	printf ("Device-ID: %08lX siliconRev:%d bootloader-version:%06X\n", deviceId, siliconRev, bootloaderVersion);
	const unsigned long expectedDeviceId = 0x04C81193;
	if (deviceId == expectedDeviceId)
	{
	    printf ("Device is a CY8C4245AXI-483\n");
	}
    }

    int firstFreeRowNumber = 0;
    int lastFreeRowNumber = 0;
    if (PsocGetFlashSize(0, &firstFreeRowNumber, &lastFreeRowNumber) < 0)
    {
	return 1;
    }
    printf ("firstRowNumber:%d lastRowNumber:%d\n", firstFreeRowNumber, lastFreeRowNumber);

    //---
    //--- load file and display overview
    //---

    CyacdFileContent data;
    if (argc > 2)
    {
	const char *filename = argv[2];
	if (PsocFileLoadCyacd(filename, data) == 0)
	{
	    PsocFileDumpContents(data, 1);
	}
	else
	{
	    fprintf(stderr, "failed to load file %s\n", filename);
	}
    }

    //---
    //--- check consistency between chip and file
    //---

    if ((deviceId != data.siliconId) || (siliconRev != data.siliconRev))
    {
	fprintf (stderr, "chip to program does not match chip if file\n");
	fprintf (stderr, "  chip: %08lX rev %d\n", deviceId, siliconRev);
	fprintf (stderr, "  file: %08lX rev %d\n", data.siliconId, data.siliconRev);
	return 1;
    }

    if ((data.firstRowToProgram < firstFreeRowNumber) || (data.lastRowToProgram > lastFreeRowNumber))
    {
	fprintf (stderr, "rows to program does not fit into device\n");

	if (data.firstRowToProgram < firstFreeRowNumber)
	{
	    fprintf (stderr, " firstFreeRowNumber: %d\n firstRowToProgram: %d\n",
		     firstFreeRowNumber,
		     data.firstRowToProgram);
	}

	if (data.lastRowToProgram > lastFreeRowNumber)
	{
	    fprintf (stderr, " lastFreeRowNumber: %d\n lastRowToProgram: %d\n",
		     lastFreeRowNumber,
		     data.lastRowToProgram);
	}
	return 1;
    }

    if (PsocSyncBootloader() != 0)
    {
	fprintf (stderr, "failed to sync bootloader for programming\n");
	return 1;
    }

    //---
    //--- Write file content to chip
    //---

    if (PsocProgramFileData(data))
    {
	fprintf (stderr, "programming failed\n");
	return 1;
    }

    //---
    //--- exit bootloader and start application after successfully programming device.
    //---
    PsocExitBootloader();

    return 0;
}
