/*
    uart programmer for Cypress PSoC.
    Input from and output to the serial port of the PSoC bootloader.
    Copyright (C) 2015  Michael Stickel (michael@cubic.org)

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "psoc_uart_programmer_io.h"

#include <errno.h>
#include <fcntl.h>
#include <stdio.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <unistd.h>

static void log_serial(FILE *_f, const char * _prompt, const unsigned char * _txdata, const int _txsize)
{
    static const char * last_prompt = 0;
    if ((last_prompt==0) || (strcmp(last_prompt, _prompt) != 0))
    {
	fprintf (_f, "\n%s:", _prompt);
	last_prompt = _prompt;
    }

    int i;
    for (i = 0; i < _txsize; ++i)
    {
	fprintf(_f, " %02X", _txdata[i]);
    }
}

static int g_psoc_fd = -1;
static FILE * uart_log_fd = 0;

int psoc_write(const unsigned char * _buffer, const int _size)
{
    int timeout = 1000;
    int remain = _size;
    const unsigned char * p = _buffer;
    while (remain > 0)
    {
	const int n = write(g_psoc_fd, p, _size);
	if (((n < 0) && (errno==EAGAIN)) || (n == 0))
	{
	    usleep(1000);
	    if (--timeout < 0)
	    {
		return -1;
	    } 
	}
	else if (n < 0)
	{
	    return n;
	}
	else
	{
	    if (uart_log_fd)
	    {
		log_serial(uart_log_fd, "TX", p, n);
	    }
	    timeout = 1000;
	    remain -= n;
	    p += n;
	}
    }
    return _size;
}

int psoc_read(unsigned char * _buffer, const int _size)
{
    int timeout = 1000;
    int remain = _size;
    unsigned char * p = _buffer;
    while (remain > 0)
    {
	const int n = read(g_psoc_fd, p, remain);
	if (((n < 0) && (errno==EAGAIN)) || (n == 0))
	{
	    if (--timeout < 0)
	    {
		return -1;
	    }
	    usleep(1000);
	}
	else if (n < 0)
	{
	    printf ("err:%d\n", errno);
	    perror ("read");
	    return n;
	}
	else
	{
	    if (uart_log_fd)
	    {
		log_serial(uart_log_fd, "RX", p, n);
	    }
	    timeout = 1000;
	    remain -= n;
	    p += n;
	}
    }
    return _size;
}

int open_psoc(const char * _deviceFilename)
{
    g_psoc_fd = open(_deviceFilename, O_RDWR | O_NONBLOCK);
    if (g_psoc_fd < 0)
    {
	fprintf(stderr, "failed to open serial port %s\n", _deviceFilename);
	return -1;
    }

    uart_log_fd = fopen("uart.log", "w");
    return 0;
}
